import struct

def function_basics():
    def euclidean_distance(x1, y1, z1, x2, y2, z2):
        return ((x1 - x2) ** 2 + (y1 - y2) ** 2 + (z1 - z2) ** 2) ** 0.5
    
    def only_side_effect():
        print("This function has a side effect.")
    
    def function_with_return():
        print("This function has a side effect and returns a value.")

        return "return values can be any type!"
    
    ret1 = function_with_return()
    ret2 = only_side_effect()

    print(ret1)
    print(ret2)

    def distance(x, y=0, *, pow, show_formula=False):
        dist = abs(x - y) ** pow

        if show_formula:
            # display rounds to 2 decimal places
            print(f"||{x} - {y}|| ** {pow} = {dist:0.2f}")

        return dist
    
    try:
        print(distance(1, 4))
    except TypeError as e:
        print(e)

    try:
        print(distance(1, 4, 2))
    except TypeError as e:
        print(e)

    print(distance(1, 4, pow=2))

    # y defaults to  0
    print(distance(1, pow=2))
    
    # show formula is optional keyword argument
    print(distance(2, pow=3, show_formula=True))


def float_basics():
    x = 1
    print(f"x = {x} is a {type(x)}")
    x = 1.
    print(f"x = {x} is a {type(x)}")

    f = float("10.5")
    print(f, type(f))

    big = 3e6
    print(big)


def float_repr():
    def get_binary_float_repr(num):
        """
        Returns the binary representation of `num` when `num` is stored as a float.

        parameters
            - num: <float> 
        """
        rep = "".join("{:0>8b}".format(c) for c in struct.pack("!d", num))

        sgn = rep[0]
        exp = rep[1:12]
        mantissa = rep[12:]

        print(f"sign: {sgn}")
        print(f"exponent: {exp} ({int(exp, 2) - 1023}) ({int(exp, 2)}-1023)")
        print(f"mantissa: {mantissa}")

        return rep

    get_binary_float_repr(103.)
    get_binary_float_repr(0.625)


def float_equality():
    def close(value, target, epsilon=1e-8):
        return abs(value - target) < epsilon

    print(sum([0.1] * 10) == 1.0)       # should print False
    print(close(sum([0.1] * 10), 1.0))  # should print True


def tuple_basics():
    def multiple_returns(s):
        return s[0], s[-1], len(s)

    first, last, length = multiple_returns("hello world!")
    print(first, last, length)

    # print exception instead of crashing program
    # try-except will be explained in a later recitation
    try:
        tup = ("text", True, 0.4, 10, (-1, "abc"))
        tup[1] = False
    except Exception as e:
        print(e)


def list_basics():
    my_list = ["Matt", 23, 3.8, ["MEng", "TA"]]
    my_list[2] += 0.1
    print(my_list)

    my_list[3].append("Dance")
    print(my_list)

    two_dimensions = [[0, 1, 2],
                      [3, 4, 5]]
    print(two_dimensions[0])
    print(two_dimensions[1][2])


def range_examples():
    def print_range(r):
        for i in r:
            print(i, end=" ")
        print()
    
    n = 12
    print_range(range(n))
    print_range(range(1, n + 1))

    L = ["a", "b", "c"]
    print_range(range(len(L)))

    print_range(range(n, 0, -1))


def iteration_examples():
    # iterate over items
    fruits = ["apple", "banana", "orange"]
    for fruit in fruits:
        print(fruit)

    # use indices when you need to look at more than one
    # item at a time
    profit = [100, 250, 750, 7500]
    yoy_growth = []
    for i in range(1, len(profit)):
        yoy_growth.append((profit[i] - profit[i - 1]) / profit[i - 1])
    print(yoy_growth)

    # iterate over indices and values
    for rank, fruit in enumerate(fruits):
        print(f"{rank + 1}: {fruit}")

    def is_prime(n):
        for i in range(2, n):
            if n % i == 0:
                return False
        return True

    # removing items while iterating forward
    L = [2, 2, 3, 4, 5, 6, 7, 7, 8, 9, 9, 10]
    for i in L:
        if is_prime(i):
            L.remove(i)
    print(L) # doesn't work
    
    # removing items while iterating backward
    L = [2, 2, 3, 4, 5, 6, 7, 7, 8, 9, 9, 10]
    L.reverse()
    for i in L:
        if is_prime(i):
            L.remove(i)
    print(L) # doesn't work

    # removing items with indices, iterating forwards
    # L = [2, 2, 3, 4, 5, 6, 7, 7, 8, 9, 9, 10]
    # for i in range(len(L)):
    #     if is_prime(L[i]):
    #         L.pop(i) # throws IndexError
    # print(L) # doesn't work
    
    # removing items with indices, iterating backwards
    L = [2, 2, 3, 4, 5, 6, 7, 7, 8, 9, 9, 10]
    for i in range(len(L) - 1, -1, -1):
        if is_prime(L[i]):
            L.pop(i)
    print(L)


def iteration_idioms():
    def collatz(n):
        if n % 2 == 0:
            return n // 2
        return int(3 * n + 1)
    
    def count_collatz(n):
        steps = 0
        while n != 1:
            n = collatz(n)
            steps += 1
        return steps

    # idiom #1: finding the "best"
    L = [11, 22, 33, 44, 55, 66, 77, 88, 99]
    best_steps = None
    best_num = None

    for num in L:
        steps = count_collatz(num)
        if best_steps is None or steps > best_steps:
            best_steps = steps
            best_num = num
    
    print(f"Number with most steps: {best_num} ({best_steps} steps)")

    # applying a function to all elements of a list
    for steps in map(count_collatz, L):
        print(steps, end=" ")
    print()

    # iterate over two lists with zip
    for num, steps in zip(L, map(count_collatz, L)):
        print(f"{num}: {steps} steps")


def string_slicing():
    # iterate over the whole string, but stepping backwards
    s = "just a random string"
    reverse_s = s[::-1]
    print(reverse_s)

    def insert(orig, new, i):
        return orig[:i] + new + orig[i:]
    
    print(insert("TATATATA", "CGCGCG", 4))


def list_slicing():
    # make a shallow copy
    L = [1, 2, 3]
    L_copy = L[:]
    L_copy[0] = -1
    print(L_copy, L)

    # make a shallow copy while reversing the list
    L = [1, 2, 3]
    L_reverse = L[::-1]
    print(L_reverse, L)

    # shallow copy doesn't work if we have mutable objects in our list
    L = [[1, 2, 3], [4, 5, 6]]
    L_copy = L[:]
    L_copy[0][0] = -1
    print(L_copy, L)


def bisection_search():
    def bsearch(Q, L):
        lo = 0
        hi = len(L) - 1

        test = (lo + hi) // 2
        while lo < hi:
            if Q(L[test]): # answer could be smaller
                hi = test
            else: # answer is definitely larger
                lo = test + 1
            test = (lo + hi) // 2
        
        return L[test]
    
    def query1(x):
        return x ** 2 > 789

    # finds the first integer n such that n ** 2 > 125
    print(bsearch(query1, list(range(1000))))


if __name__ == "__main__":
    # function_basics()
    # float_basics()
    # float_repr()
    # float_equality()
    # tuple_basics()
    # list_basics()
    # range_examples()
    # iteration_examples()
    # iteration_idioms()
    # string_slicing()
    # list_slicing()
    # bisection_search()
    pass
